// C Source File
// Created 2/9/01; 10:52:07 AM
//Original code from: Sample TSR by Greg Dietsche - Last Modification: 10/22/2002
//Modified by: Thomas Naeff, 16.03.2006


//-----------------------------------------------------------------------------
// Undocumented macro in TIGCC v0.95 and above: Do not create global imports
// for startup code. This macro is defined exclusively for ttstart; do not use
// it in your own programs. In ttstart, it only suppresses a lot of warnings.
//-----------------------------------------------------------------------------
#define _NO_INCLUDE_PATCH
#define MIN_AMS 200

#define fatal(s) {ST_showHelp(s);return;}

//Seconds to wait for password
#define seconds 10

//TODO: Set Your Event Hook's Name
//Hook Name: Must Be 8 non-zero characters! Use Spaces if necessary to fill this requirement
//You should edit this to reflect the name of your event hook...
//When uninstalling your event hook, the uninstall program will use this name in its list of
//installed TSR Event Hooks
	#define TSR_NAME "tsrPASS "

//Usually this results in smaller code, OPTIMIZE_ROM_CALLS also makes calling rom functions faster
//than they would be normally
	#define OPTIMIZE_ROM_CALLS

//TODO: Do you need to have TIGCC detect the calc type?
//For Sample TSR, defining NO_CALC_DETECT will result in smaller code
//You should comment this out to see if smaller code is produced in your own projects...
//Often times, undefining (commenting out) NO_CALC_DETECT will result in
//smaller code if you use the so-called psuedo constants such as KEY_UP in compat.h
	#define NO_CALC_DETECT

//TODO: Select the calculators that you are compiling for...
//In the project otions, maybe you should uncheck all checkmarks at
//Project -> Options... -> Compilation -> Calculator
//(If these project options are available in your version of TIGCC)
	#define USE_TI89
	#define USE_TI92PLUS
	#define USE_V200

//Necessary Header Files
	#include <tigcclib.h>
	#include <kbd.h>

//TODO: Add your own header files, and other information that needs a global scope
	void Install_TSR(void);
	void EventHandler(EVENT *ev);
	void Install(void);
	void settings(void);
	void lock(void);
	short passInput(const char *message, char *buffer,short timeout,short length,short hide);
	short memprotect(void);
	char password[8];
	short secprot=TRUE,diaprot=FALSE,APDprot=FALSE,MEMprot=FALSE;
	short APDSecs = 60;
	short blink=500;
	void resetAPD(void);
	void resetblink(void);

//This Code needs TIGCC 0.94 beta 19 at the very least to function properly
//Thomas: I set this to TIGCC 0.96 beta 6 because I tested it with this version.
#if __TIGCC__== 0 && __TIGCC_MINOR__< 96 && __TIGCC_BETA__ < 6
	#warning Sample TSR - Old TIGCC Version -- Check tigcc.ticalc.org for updates
#endif

///////////////////////
// EV_hook Header		//
/////////////////////
//The Total Length of this header is 22 bytes
//The Offset to jmp Install_TSR:l is 16 bytes
//The Offset to gpOldHandler is 12 bytes
//Here is now the patch for TIGCC 0.94 from Kevin Kovler (added by Thomas Naeff)
asm(".section _stl1 | First library startup section. This is what the executable starts with.
		.ascii \"evHk" TSR_NAME "\"
		.globl gpOldHandler
		gpOldHandler:
		.long 0
		__jmp__Event_Hook__:
		jmp Install_TSR:l

		.global __calculator
		__calculator:
		.word 0");

	//Final offset: 22 bytes
///////////////////////////////
// End of EV_hook Header		//
/////////////////////////////

//Global Variables Go Here
extern EVENT_HANDLER gpOldHandler; 		//a pointer to a pointer to the old handler
//Example:
//char UserInput[20];
//int InputCount;

//Install Function -  this function will be executed ONCE, and ONLY ONCE
void Install_TSR(void)
{
		extern char __jmp__Event_Hook__;
		char *ptr=&__jmp__Event_Hook__;

	//Set EventHandler as the default entry point from now on
	//instead of Install_TSR (this function)
		*(unsigned long*)&ptr[2]=(unsigned long)EventHandler;

	#ifndef NO_CALC_DETECT
		asm volatile("
		/* Start Of Calculator Detection Code -- From TIGCC Sources */
				move.l 0xC8,%%a0
				/* Calculator Detection */
				moveq #1,%%d0
				move.l %%a0,%%d1
				and.l #0x400000,%%d1
				jbne __calc_in_d0__
				clr.w %%d0
				move.l (%%a0,0x2F*4),%%a1 /* ScrRect */
				cmp.b #200,2(%%a1)
				jbcs __calc_in_d0__
				moveq #3,%%d0
			__calc_in_d0__:
				lea __calculator(%%pc),%%a1
				move.w %%d0,(%%a1)
		/* End Of Calculator Detection Code -- From TIGCC Sources */"
		:
		:
		: "a0", "a1", "d1");
	#endif

		Install(); //Code run on first install of tsrPASS
}

void EventHandler(EVENT *ev)
{
//////////////////////////////////////////////
//Prepatory Setup for the Event Handler    //
////////////////////////////////////////////

// a special variable: pointer to a function
	EVENT_HANDLER OldHook;

	#ifdef OPTIMIZE_ROM_CALLS
		void *__saved_a5__;//a variable to save the register a5 so it can be restored later
		asm("/*START: Support for OPTIMIZE_ROM_CALLS*/");
		asm("move.l %%a5,%0":"=p"(__saved_a5__));
		asm("move.l 0xC8,%a5");
		asm("/*END: Support for OPTIMIZE_ROM_CALLS*/");
	#endif

/////////////////////////////////////////////////////////////////

//////////////////////////////////////////
//Your Event Handler Code Starts Here  //
////////////////////////////////////////

//TODO: write your very own event hook code
	if(ev->Type==CM_KEYPRESS)
	{
		if(ev->extra.Key.Code==KEY_OFF){
			if(secprot){
				lock();
				ev->Type=CM_ON;
			}else{
				off();
				ev->Type=CM_ON;
			}
			resetAPD();
		}
		if(ev->extra.Key.Code==16651){
			if(diaprot){
				lock();
				ev->Type=CM_NOTHING;
			}else{
				off();
				ev->Type=CM_NOTHING;
			}
			resetAPD();
		}
		if(ev->extra.Key.Code==KEY_MEM&&MEMprot){
			if(!memprotect()){
				ev->Type=CM_NOTHING;
				fatal("MEMORY BLOCKED");
			}
			resetAPD();
		}
		if(ev->extra.Key.Code==16647){
			settings();
			ev->Type=CM_NOTHING;
			resetAPD();
			resetblink();
		}
	}
	if(OSTimerCurVal(APD_TIMER)==1){
		if(APDprot){
			lock();
		}else{
			off();
		}
		ev->Type=CM_NOTHING;
		resetAPD();
	}

///////////////////////////////
//End of Your Event Handler //
/////////////////////////////

///////////////////////////////////////
//Call the Old hook if there was one//
/////////////////////////////////////

//get the address of the next TSR in line to be executed NULL if there is none
  OldHook=gpOldHandler;

//Call the old hook if there is one to be called
  if(OldHook)
  {
  	asm("								/*This inline asm takes the place of OldHook(ev);
  												the difference is that a2 will also contain ev
  												as a parameter to support TEOS. Without this code,
  												you would need to use the switch -ffixed a2 when
  												compiling this example.*/
  	move.l %0,%%a2			/*TEOS expects the parameter to be in a2*/
  	move.l %%a2,-(%%sp)	/*other event hooks expect it on the stack*/
  	move.l %1,%%a0			/*Get the address to call*/
  	jsr (%%a0)					/*Call the next event hook*/
  	addq.l #4,%%sp			/*restore the stack*/
  	"::"p"(ev),"g"(OldHook):"a0","a2");
  }//end if(OldHook)

	#ifdef OPTIMIZE_ROM_CALLS
		asm("/*START: Support for OPTIMIZE_ROM_CALLS*/");
		asm("move.l %0, %%a5	/*support for OPTIMIZE_ROM_CALLS*/"::"p"(__saved_a5__));
		asm("/*END: Support for OPTIMIZE_ROM_CALLS*/");
	#endif

}

///////////////////////////////////////////////////////////////////////////
//A Small function to prove that it is possible to call other functions //
/////////////////////////////////////////////////////////////////////////
void Install(void){
	char temp[8];
	passInput("New Password:",password,0,8,1);
	passInput("Confirm Password:",temp,0,8,1);
	while(memcmp(password,temp,8)!=0){
		passInput("Retry Password:",password,0,8,1);
		passInput("Confirm:",temp,0,8,1);
	}
	resetAPD();
	resetblink();
}

short passInput(const char *message, char *buffer,short timeout,short passwidth,short hide){
	GKeyFlush();
	char oneKey[1];
	int index;
	int length = 0;
	LCD_BUFFER screen;
	LCD_save(&screen);
	unsigned long int timeleft = seconds*20;
	short input;
	SetAlphaStatus(0);
	//clrscr();
	int width = DrawStrWidth(message,F_6x8)+10;
	if(width<passwidth*10+10)width=passwidth*10+10;
	int x1 = (ScrRect->xy.x1-width)/2;
	int x2 = x1+width;
	int top = ScrRect->xy.y1/2-15;
	int bottom = ScrRect->xy.y1/2+15;
	int start = x1+(width-(10*passwidth-5))/2;
	DrawLine(x1,top,x2,top,A_THICK1);
	DrawLine(x1,bottom,x2,bottom,A_THICK1);
	DrawLine(x1,top,x1,bottom,A_THICK1);
	DrawLine(x2,top,x2,bottom,A_THICK1);
	for(index=top+1;index<bottom;index++){
		DrawLine(x1+1,index,x2-1,index,A_REVERSE);
	}
	DrawStr(x1+(width-DrawStrWidth(message,F_6x8))/2,top+5,message,A_NORMAL);
	for(index=0;index<8;index++){
		buffer[index] = 0;
	}
	while(1){
		if(kbhit()){
			input=ngetchx();
			timeleft = 50000;
			if((input==KEY_BACKSPACE)){
				if(length>0){
					length--;
					buffer[length] = 0;
				}
			}else if(input==KEY_ENTER){
				LCD_restore(&screen);
				return 0;
			}else if(input==KEY_ESC){
				for(index=0;index<passwidth;index++){
					buffer[index]=0;
				}
				LCD_restore(&screen);
				return 1;
			}else{
				if(length<passwidth){
					buffer[length] = (char)input;
					length++;
				}
			}
			for(index=0;index<passwidth;index++){
				if(index < length){
					if(hide){
						DrawStr(start+index*10,bottom-12,"*",A_NORMAL);
					}else{
						sprintf(oneKey,"%c",buffer[index]);
						DrawStr(start+index*10,bottom-12,oneKey,A_NORMAL);
					}
				}else{
					DrawStr(start+index*10,bottom-12," ",A_REPLACE);
				}
			}
		}else if(timeout){
			timeleft--;
			if(timeleft==0){
				for(index=0;index<passwidth;index++){
					buffer[index]=0;
				}
				LCD_restore(&screen);
				return 1;
			}
		}
		idle();
	}
}

void lock(void){
	char temp[8];
	short bad;
	GKeyFlush();
	while(1){
		//clrscr();
		off();
		bad = passInput("Enter Password:",temp,1,8,1);
		if(!bad && memcmp(password,temp,8)==0)break;
	}
	fatal("tsrPass")
}

void settings(void){
	char temp1[8],temp2[8];
	LCD_BUFFER screen;
	int cont=1;
	if(passInput("Current Password:",temp1,0,8,1))return;
	if(memcmp(password,temp1,8)!=0){
		fatal("WRONG PASSWORD!");
	}
	LCD_save(&screen);
	while(cont){
		clrscr();
		printf("\n1. New Pass ->       ");
		printf("\n2. 2ND-ON Protect  %d",secprot);
		printf("\n3. 3RD-ON Protect  %d",diaprot);
		printf("\n4. APD Protect     %d",APDprot);
		printf("\n5. MEM Protect     %d",MEMprot);
		printf("\n6. APD Time ->     %d",APDSecs);
		printf("\n7. Cursor Blink -> %d",blink);
		printf("\nESC. EXIT          ");
		switch(ngetchx()){
			case '1':
				if(passInput("New Password:",temp1,0,8,1))break;
				if(passInput("Confirm Password:",temp2,0,8,1))break;
				if(memcmp(temp1,temp2,8)!=0){
					ST_helpMsg("No Match!");
					break;
				}
				memcpy(password,temp1,8);
				ST_helpMsg("Password set!");
				break;
			case '2':
				secprot=!secprot;
				break;
			case '3':
				diaprot=!diaprot;
				break;
			case '4':
				APDprot=!APDprot;
				break;
			case '5':
				MEMprot=!MEMprot;
				break;
			case '6':
				if(passInput("APD",temp1,0,3,0))break;
				if(!isdigit(temp1[0]))break;
				if(!isdigit(temp1[1])){
					APDSecs=temp1[0]-'0';
					break;
				}
				if(!isdigit(temp1[2])){
					APDSecs=10*(temp1[0]-'0')+temp1[1]-'0';
					break;
				}else{
					APDSecs=100*(temp1[0]-'0')+10*(temp1[1]-'0')+temp1[2]-'0';
					break;
				}
			case '7':
				if(passInput("Blink (ms):",temp1,0,3,0))break;
				if(!isdigit(temp1[0]))break;
				if(!isdigit(temp1[1])){
					blink=temp1[0]-'0';
					break;
				}
				if(!isdigit(temp1[2])){
					blink=10*(temp1[0]-'0')+temp1[1]-'0';
					break;
				}else{
					blink=100*(temp1[0]-'0')+10*(temp1[1]-'0')+temp1[2]-'0';
					break;
				}
			case KEY_ESC:
				cont=0;
				break;
		}
	}
	LCD_restore(&screen);
}

short memprotect(void){
	char temp[8];
	passInput("Enter Password:",temp,0,8,1);
	return memcmp(password,temp,8)==0;
}

void resetAPD(void){
	OSFreeTimer(APD_TIMER);
	OSRegisterTimer(APD_TIMER,APDSecs*20);
}

void resetblink(void){
	OSFreeTimer(CURSOR_TIMER);
	OSRegisterTimer(CURSOR_TIMER,blink/50);
}